#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import time
from multiprocessing import Process
from queue import Empty

from flask import Flask, request
from gevent.pywsgi import WSGIServer
from setproctitle import setproctitle

from healthcheck_python.release import __version__


class HealthCheckApi(Process):
	"""
	API responder class
	Creates a flask instance and reports the health status
	"""

	def __init__(self, host, port, status_queue, daemon=False):
		super().__init__()

		self._host = host
		self._port = port
		self._status_queue = status_queue
		self.daemon = daemon

		self._app = Flask(__name__)

		self._app.add_url_rule('/', 'index', view_func=HealthCheckApi._index)
		self._app.add_url_rule('/health', 'health', view_func=self._health)

		self._http_server = WSGIServer((self._host, self._port), self._app)

	def __del__(self):
		self._http_server.close()
		self.terminate()

	def run(self):
		setproctitle(self.__class__.__name__)
		self._http_server.serve_forever()

	@staticmethod
	def _index():
		return f"Hello there! I'm healthcheck-python v{__version__}"

	def _get_status(self) -> dict:
		"""
		Get a single valid message from queue
		:return: dict
		"""
		while True:
			try:
				status = self._status_queue.get(block=True, timeout=1)
				if time.time() - status[0] <= 0.5:
					status = status[1]
					break
			except Empty:
				status = {'status': False, 'services': {}}
				break

		return status

	def _ready(self):
		"""
		Readiness check path
		/health
		:return: overall readiness str(boolean).
		:return: If verbose mode enabled, return a dict with details about every service
		"""
		is_verbose = "v" in request.args.keys()

		status_message = self._get_status()

		if is_verbose:
			return status_message

		return {'ready': status_message['ready']}

	def _health(self):
		"""
		Health check path
		/health
		:return: overall status str(boolean).
		:return: If verbose mode enabled, return a dict with details about every service
		"""
		is_verbose = "v" in request.args.keys()

		status_message = self._get_status()

		if is_verbose:
			return status_message

		return {'status': status_message['status']}



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
import time

import healthcheck_python.config as config
from healthcheck_python.service.periodic_service import PeriodicService
from healthcheck_python.utils.pipeline import start
from healthcheck_python.utils.utils import ServiceStatus, ServiceOperation


def periodic(_func=None, *, service='', timeout=5):
	"""
	Periodic check decorator
	This decorator only defines a new periodic service
	:param _func: Wrapped function
	:param service: Service name. This name will be reported with API call.
	If it's empty, class name, if wraps a class, or function name, if wraps a function, will be used
	:param timeout: The timeout in seconds needed between to consecutive _func() calls
	before marking the service down
	:return: original return values of _func()
	"""
	start()

	def wrapper(func):
		@functools.wraps(func)
		def wrapper_func(*args, **kwargs):
			service_name = service
			if service_name == '':
				service_name = func.__qualname__.split(".")[0]
			ret_val = func(*args, **kwargs)
			if not config.TEST_MODE:
				config.message_queue.put(
					{
						'type': PeriodicService, 'op': ServiceOperation.CREATE,
						'name': service_name, 'timeout': timeout
					}
				)

			return ret_val

		return wrapper_func

	if _func is None:
		return wrapper

	return wrapper(_func)


def healthy(_func=None, *, service=''):
	"""
	Periodic health check decorator
	Add this to your periodically called functions to report a service healthy
	:param _func: Wrapped function
	:param service: Service name. This name will be reported with API call
	If it's empty, class name, if wraps a class, or function name, if wraps a function, will be used
	The service has to be already defined before healthy is called
	:return: original return values of _func()
	"""
	start()

	def wrapper(func):
		@functools.wraps(func)
		def wrapper_func(*args, **kwargs):
			service_name = service
			if service_name == '':
				service_name = func.__qualname__.split(".")[0]
			ret_val = func(*args, **kwargs)
			end_time = time.time()
			if not config.TEST_MODE:
				config.message_queue.put(
					{
						'name': service_name, 'op': ServiceOperation.ADD_HEALTH_POINT, 'end_time': end_time
					}
				)

			return ret_val

		return wrapper_func

	if _func is None:
		return wrapper

	return wrapper(_func)


def fps(_func=None, *, service=''):
	"""
	FPS Calculation decorator
	Add this to your periodically called functions to calculate the fps
	:param _func: Wrapped function
	:param service: Service name. This name will be reported with API call
	If it's empty, class name, if wraps a class, or function name, if wraps a function, will be used
	The service has to be already defined before healthy is called
	:return: original return values of _func()
	"""

	start()

	def wrapper(func):
		@functools.wraps(func)
		def wrapper_func(*args, **kwargs):
			service_name = service
			if service_name == '':
				service_name = func.__qualname__.split(".")[0]
			start_time = time.time()
			ret_val = func(*args, **kwargs)
			end_time = time.time()
			if not config.TEST_MODE:
				config.message_queue.put(
					{
						'name': service_name, 'op': ServiceOperation.ADD_FPS_POINT,
						'start_time': start_time, 'end_time': end_time
					}
				)

			return ret_val

		return wrapper_func

	if _func is None:
		return wrapper

	return wrapper(_func)


def mark_ready(_func=None, *, service=''):
	"""
	Mark a service ready to serve
	Also clears done flag
	:param _func: Wrapped function
	:param service: Service name. This name will be reported with API call
	If it's empty, class name, if wraps a class, or function name, if wraps a function, will be used
	The service has to be already defined before healthy is called
	:return: original return values of _func()
	"""

	start()

	def wrapper(func):
		@functools.wraps(func)
		def wrapper_func(*args, **kwargs):
			service_name = service
			if service_name == '':
				service_name = func.__qualname__.split(".")[0]
			ret_val = func(*args, **kwargs)
			if not config.TEST_MODE:
				config.message_queue.put(
					{
						'name': service_name, 'op': ServiceOperation.MARK_READY,
						'status': ServiceStatus.READY
					}
				)

			return ret_val

		return wrapper_func

	if _func is None:
		return wrapper

	return wrapper(_func)


def mark_done(_func=None, *, service=''):
	"""
	Mark a service done and make it successful indefinitely
	:param _func: Wrapped function
	:param service: Service name. This name will be reported with API call
	If it's empty, class name, if wraps a class, or function name, if wraps a function, will be used
	The service has to be already defined before healthy is called
	:return: original return values of _func()
	"""

	start()

	def wrapper(func):
		@functools.wraps(func)
		def wrapper_func(*args, **kwargs):
			service_name = service
			if service_name == '':
				service_name = func.__qualname__.split(".")[0]
			ret_val = func(*args, **kwargs)
			if not config.TEST_MODE:
				config.message_queue.put(
					{
						'name': service_name, 'op': ServiceOperation.MARK_DONE,
						'status': ServiceStatus.DONE
					}
				)

			return ret_val

		return wrapper_func

	if _func is None:
		return wrapper

	return wrapper(_func)
